/* /////////////////////////////////////////////////////////////////////////////
 * File:		perf_test.cpp
 *
 * Purpose:		Test program for the auto_buffer template.
 *
 * Created		
 * Updated:		 17th February 2004
 *
 * Copyright:	Synesis Software Pty Ltd, (c) 2003. All rights reserved.
 *
 * Home:		www.synesis.com.au/software
 *
 * ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef __STLSOFT_COMPILER_IS_MSVC
# pragma warning(disable : 4244) // Since VC++ 6 does not listen to the static_cast
# pragma warning(disable : 4711) // Don't need to know "function '. . .' selected for automatic inline expansion"
#endif /* __STLSOFT_COMPILER_IS_MSVC */

/* ////////////////////////////////////////////////////////////////////////// */

#define SUCCINCT_OUTPUT_DDJ_0504


#include <stdio.h>						//	fprintf
#include <stdlib.h>						//	atoi, alloca()
#include <malloc.h>						//	malloc & free, _alloca()
#include <new>							//	new & delete

#include <stlsoft.h>

typedef stlsoft_ns_qual(ss_byte_t)				element_t;

#include <stlsoft_auto_buffer.h>		//	stlsoft's auto_buffer

stlsoft_ns_using(auto_buffer)

#include <winstl_processheap_allocator.h>

typedef winstl_ns_qual(processheap_allocator)<element_t>	ator_t;



#if defined(WIN32) || \
    defined(_WIN32)

#if defined(__STLSOFT_COMPILER_IS_BORLAND)
# include <winstl_tick_counter.h>

 typedef winstl_ns_qual(tick_counter)			performance_counter;
#else
# include <winstl_performance_counter.h>

 typedef winstl_ns_qual(performance_counter)	performance_counter;
#endif /* compiler */

#else // Otherwise assume UNIX

# include <unixstl_performance_counter.h>

 typedef unixstl_ns_qual(performance_counter)	performance_counter;

#endif /* WIN32 || _WIN32 */

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef __STLSOFT_COMPILER_IS_MSVC
# pragma warning(disable : 4244) // Since VC++ 6 does not listen to the static_cast
# pragma warning(disable : 4711) // Don't need to know "function '. . .' selected for automatic inline expansion"
#endif /* __STLSOFT_COMPILER_IS_MSVC */

/* /////////////////////////////////////////////////////////////////////////////
 * vector?
 */

#ifdef TEST_VECTOR
# undef TEST_VECTOR
#endif /* TEST_VECTOR */

#if defined(__STLSOFT_COMPILER_IS_BORLAND) || \
    defined(__STLSOFT_COMPILER_IS_COMO) || \
    defined(__STLSOFT_COMPILER_IS_DMC) || \
    defined(__STLSOFT_COMPILER_IS_GCC) || \
    defined(__STLSOFT_COMPILER_IS_INTEL) || \
    defined(__STLSOFT_COMPILER_IS_MWERKS) || \
    defined(__STLSOFT_COMPILER_IS_MSVC)
# define TEST_VECTOR
#endif /* compiler */

#ifdef TEST_VECTOR
# include <vector>

 stlsoft_ns_using_std(vector)
#endif /* TEST_VECTOR */

/* /////////////////////////////////////////////////////////////////////////////
 * alloca?
 */

#ifdef TEST_ALLOCA
# undef TEST_ALLOCA
#endif /* TEST_ALLOCA */

#if defined(__STLSOFT_COMPILER_IS_BORLAND) || \
    defined(__STLSOFT_COMPILER_IS_COMO) || \
    defined(__STLSOFT_COMPILER_IS_DMC) || \
    defined(__STLSOFT_COMPILER_IS_GCC) || \
    defined(__STLSOFT_COMPILER_IS_INTEL) || \
    defined(__STLSOFT_COMPILER_IS_MWERKS) || \
    defined(__STLSOFT_COMPILER_IS_MSVC) || \
    defined(__STLSOFT_COMPILER_IS_WATCOM)
# define TEST_ALLOCA
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * VLA?
 */

#ifdef TEST_VLA
# undef TEST_VLA
#endif /* TEST_VLA */

#if /* defined(__STLSOFT_COMPILER_IS_COMO) || \
     */defined(__STLSOFT_COMPILER_IS_DMC) || \
    defined(__STLSOFT_COMPILER_IS_GCC)/*  || \
    defined(__STLSOFT_COMPILER_IS_WATCOM) */
# define TEST_VLA
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * printf() format string
 */

#if defined(__STLSOFT_COMPILER_IS_BORLAND) || \
    defined(__STLSOFT_COMPILER_IS_COMO) || \
    defined(__STLSOFT_COMPILER_IS_INTEL) || \
    defined(__STLSOFT_COMPILER_IS_MSVC) || \
    defined(__STLSOFT_COMPILER_IS_VECTORC) || \
    defined(__STLSOFT_COMPILER_IS_WATCOM)
# define test_printf_fmt	"%I64d"
#else
# define test_printf_fmt	"%lld"
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants and definitions
 */

const int		WARMUPS						=	2;
#if 0
const size_t	AUTO_BUFFER_INTERNAL_SIZE	=	64;
#else
const size_t	AUTO_BUFFER_INTERNAL_SIZE	=	512;
#endif /* 0 */

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

typedef union Automatic
{
	element_t	_100[100];
	element_t	_1000[1000];
	element_t	_10000[10000];

} Automatic;

/* /////////////////////////////////////////////////////////////////////////////
 * Globals
 */

//extern volatile char	**s_p	=	0;

/* /////////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

int test_automatic(size_t alloc_size, int iterations);
int test_malloc(size_t alloc_size, int iterations);
int test_new(size_t alloc_size, int iterations);
int test_alloca(size_t alloc_size, int iterations, bool bTryAllocA);
int test_vector(size_t alloc_size, int iterations);
int test_vla(size_t alloc_size, int iterations, bool bVLA);
int test_auto_buffer(size_t alloc_size, int iterations);

int main(int argc, char **argv)
{
	int	iRet = 0;

#ifdef SUCCINCT_OUTPUT_DDJ_0504
	fprintf(stdout, "auto_buffer (%s):\t", __STLSOFT_COMPILER_LABEL_STRING);
#else /* ? SUCCINCT_OUTPUT_DDJ_0504 */
	fprintf(stdout, "auto_buffer<> test: %s\n", __STLSOFT_COMPILER_LABEL_STRING);
#endif /* SUCCINCT_OUTPUT_DDJ_0504 */

	if(argc < 3)
	{
usage:
		fprintf(stderr, "  USAGE: Listing4 [-a] [-v] <alloc-size> <iterations>\n");
		fprintf(stderr, "    -a              -   do an alloca() test (likely to\n");
		fprintf(stderr, "                    -     crash with insufficient stack)\n");
		fprintf(stderr, "    -v              -   do a VLA test (likely to\n");
		fprintf(stderr, "                    -     crash with insufficient stack)\n");
		fprintf(stderr, "    <alloc-size>    -   the amount of memory to allocate\n");
		fprintf(stderr, "    <iterations>    -   number of times to execute the test loop\n");

		iRet = 1;
	}
	else
	{
		char const	*pcszAllocSize	=	NULL;
		char const	*pcszIterations	=	NULL;
		bool		bTryAllocA		=	false;
		bool		bVLA			=	false;

		for(int i = 1; i < argc; ++i)
		{
			char const	*arg	=	argv[i];

			if(arg[0] == '-')
			{
				switch(arg[1])
				{
					case	'a':
						bTryAllocA = true;
						break;
					case	'v':
						bVLA = true;
						break;
					default:
						goto usage;
				}
			}
			else
			{
				if(NULL == pcszAllocSize)
				{
					pcszAllocSize = arg;
				}
				else if(NULL == pcszIterations)
				{
					pcszIterations = arg;
				}
			}
		}

#ifdef _DEBUG
		puts(pcszAllocSize);
		puts(pcszIterations);
#endif /* _DEBUG */

		size_t	alloc_size	=	static_cast<size_t>(atoi(pcszAllocSize));
		int		iterations	=	atoi(pcszIterations);

#ifndef SUCCINCT_OUTPUT_DDJ_0504
		fprintf(stdout, "internal size: %u; block size: %u\n", AUTO_BUFFER_INTERNAL_SIZE, alloc_size);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */

#ifdef WIN32
		::SetThreadPriority(::GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
#endif // WIN32

		if(alloc_size > sizeof(Automatic))
		{
			fprintf(stderr, "Size of buffer allocation cannot be larger than %u for this implementation!\n", sizeof(Automatic));
		}
		else
		{
#ifndef SUCCINCT_OUTPUT_DDJ_0504
			iRet += test_automatic(alloc_size, iterations);
			iRet += test_malloc(alloc_size, iterations);
			iRet += test_new(alloc_size, iterations);
			iRet += test_vector(alloc_size, iterations);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
			iRet += test_auto_buffer(alloc_size, iterations);
#ifndef SUCCINCT_OUTPUT_DDJ_0504
			iRet += test_alloca(alloc_size, iterations, bTryAllocA);
			iRet += test_vla(alloc_size, iterations, bVLA);
#else /* ? SUCCINCT_OUTPUT_DDJ_0504 */
			STLSOFT_SUPPRESS_UNUSED(bTryAllocA);
			STLSOFT_SUPPRESS_UNUSED(bVLA);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */

#ifndef SUCCINCT_OUTPUT_DDJ_0504
			fprintf(stdout, "Total: %d\n", iRet);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
		}
	}

	return iRet;
}


/* /////////////////////////////////////////////////////////////////////////////
 * Test functions
 */

int test_automatic(size_t alloc_size, int iterations)
{
	int					i;
	int					j;
	int					iRet = 0;
	performance_counter	counter;

	fprintf(stdout, "automatic:     ");

	for(j = 0; j < WARMUPS; ++j)
	{
		counter.start();
		for(i = 0, iRet = 0; i < iterations; ++i)
		{
			// (i) Allocate
			Automatic	_buffer;
			element_t		*buffer	=	_buffer._1000;

			// (ii) Use
//			s_p = const_cast<volatile element_t**>(&buffer);

			// Help to avoid compiler optimising away our code, by touching 
			// last element in the array ...
			buffer[alloc_size - 1] += static_cast<element_t>(i);
			// ... and testing the next to last one
			iRet ^= buffer[alloc_size - 1];

			// (iii) Deallocate
			; // Nothing to do
		}
		counter.stop();
	}

	fprintf(stdout, " \t" test_printf_fmt "\n", counter.get_milliseconds());

	return iRet;
}

int test_malloc(size_t alloc_size, int iterations)
{
	int					i;
	int					j;
	int					iRet = 0;
	performance_counter	counter;

	fprintf(stdout, "malloc():      ");

	for(j = 0; j < WARMUPS; ++j)
	{
		counter.start();
		for(i = 0, iRet = 0; i < iterations; ++i)
		{
			// (i) Allocate
			element_t	*buffer	=	(element_t*)malloc(alloc_size);

			// (ii) Use
//			s_p = const_cast<volatile element_t**>(&buffer);

			// Help to avoid compiler optimising away our code, by touching 
			// last element in the array ...
			buffer[alloc_size - 1] += static_cast<element_t>(i);
			// ... and testing the next to last one
			iRet ^= buffer[alloc_size - 1];

			// (iii) Deallocate
			free(buffer);
		}
		counter.stop();
	}

	fprintf(stdout, " \t" test_printf_fmt "\n", counter.get_milliseconds());

	return iRet;
}

int test_new(size_t alloc_size, int iterations)
{
	int					i;
	int					j;
	int					iRet = 0;
	performance_counter	counter;

	fprintf(stdout, "operator new:  ");

	for(j = 0; j < WARMUPS; ++j)
	{
		counter.start();
		for(i = 0, iRet = 0; i < iterations; ++i)
		{
			// (i) Allocate
			element_t	*buffer	=	new element_t[alloc_size];

			// (ii) Use
//			s_p = const_cast<volatile element_t**>(&buffer);

			// Help to avoid compiler optimising away our code, by touching 
			// last element in the array ...
			buffer[alloc_size - 1] += static_cast<element_t>(i);
			// ... and testing the next to last one
			iRet ^= buffer[alloc_size - 1];

			// (iii) Deallocate
			delete [] buffer;
		}
		counter.stop();
	}

	fprintf(stdout, " \t" test_printf_fmt "\n", counter.get_milliseconds());

	return iRet;
}

int test_alloca(size_t alloc_size, int iterations, bool bTryAllocA)
{
#ifdef TEST_ALLOCA
	if(!bTryAllocA)
#endif /* !TEST_ALLOCA */
	{
		return fprintf(stdout, "alloca():       -\n");
	}

#ifdef TEST_ALLOCA
	int					i;
	int					j;
	int					iRet = 0;
	performance_counter	counter;

	fprintf(stdout, "alloca():      ");
	fflush(stdout);

	for(j = 0; j < WARMUPS; ++j)
	{
		counter.start();
		for(i = 0, iRet = 0; i < iterations; ++i)
		{
			// (i) Allocate
#ifdef _MSC_VER
			element_t	*buffer	=	(element_t*)_alloca(alloc_size);
#elif defined(__STLSOFT_COMPILER_IS_GCC)
			element_t	*buffer	=	(element_t*)__builtin_alloca(alloc_size);
#else
			element_t	*buffer	=	(element_t*)alloca(alloc_size);
#endif // _MSC_VER

			// (ii) Use
//			s_p = const_cast<volatile element_t**>(&buffer);

			// Help to avoid compiler optimising away our code, by touching 
			// last element in the array ...
			buffer[alloc_size - 1] += static_cast<element_t>(i);
			// ... and testing the next to last one
			iRet ^= buffer[alloc_size - 1];

			// (iii) Deallocate
			; // Nothing to do
		}
		counter.stop();
	}

	fprintf(stdout, " \t" test_printf_fmt "\n", counter.get_milliseconds());

	return iRet;
#endif /* TEST_ALLOCA */
}

int test_vector(size_t alloc_size, int iterations)
{
	STLSOFT_SUPPRESS_UNUSED(alloc_size);
	STLSOFT_SUPPRESS_UNUSED(iterations);

#ifndef TEST_VECTOR
	return fprintf(stdout, "vector<>:      \n");
#else
#ifdef __STLSOFT_COMPILER_IS_DMC
	typedef	vector<element_t>			vector_t;
#else
	typedef	vector<element_t, ator_t>	vector_t;
#endif /* __STLSOFT_COMPILER_IS_DMC */

	int					i;
	int					j;
	int					iRet = 0;
	performance_counter	counter;

	fprintf(stdout, "vector<>:      ");

	for(j = 0; j < WARMUPS; ++j)
	{
		counter.start();
		for(i = 0, iRet = 0; i < iterations; ++i)
		{
			// (i) Allocate
			vector_t	buffer(alloc_size);

			// (ii) Use
//			s_p = const_cast<volatile element_t**>(&buffer);

			// Help to avoid compiler optimising away our code, by touching 
			// last element in the array ...
			buffer[alloc_size - 1] += static_cast<element_t>(i);
			// ... and testing the next to last one
			iRet ^= buffer[alloc_size - 1];

			// (iii) Deallocate
			; // Destructor will handle
		}
		counter.stop();
	}

	fprintf(stdout, " \t" test_printf_fmt "\n", counter.get_milliseconds());

	return iRet;
#endif /* TEST_VECTOR */
}

int test_vla(size_t alloc_size, int iterations, bool bVLA)
{
	STLSOFT_SUPPRESS_UNUSED(alloc_size);
	STLSOFT_SUPPRESS_UNUSED(iterations);
	STLSOFT_SUPPRESS_UNUSED(bVLA);

#ifdef TEST_VLA
	if(!bVLA)
#endif /* TEST_VLA */
	{
		return fprintf(stdout, "VLA:            -\n");
	}

#ifdef TEST_VLA
	int					i;
	int					j;
	int					iRet = 0;
	performance_counter	counter;

	fprintf(stdout, "VLA:           ");
	fflush(stdout);

	for(j = 0; j < WARMUPS; ++j)
	{
		counter.start();
		for(i = 0, iRet = 0; i < iterations; ++i)
		{
			// (i) Allocate
			element_t	buffer[alloc_size];

			// (ii) Use
//			s_p = const_cast<volatile element_t**>(&buffer);

			// Help to avoid compiler optimising away our code, by touching 
			// last element in the array ...
			buffer[alloc_size - 1] += static_cast<element_t>(i);
			// ... and testing the next to last one
			iRet ^= buffer[alloc_size - 1];

			// (iii) Deallocate
			; // Nothing to do
		}
		counter.stop();
	}

	fprintf(stdout, " \t" test_printf_fmt "\n", counter.get_milliseconds());

	return iRet;
#endif /* TEST_VLA */
}

int test_auto_buffer(size_t alloc_size, int iterations)
{
	typedef	auto_buffer<element_t, ator_t, AUTO_BUFFER_INTERNAL_SIZE>	buffer_t;

	int					i;
	int					j;
	int					iRet = 0;
	performance_counter	counter;

#ifndef SUCCINCT_OUTPUT_DDJ_0504
	fprintf(stdout, "auto_buffer<>: ");
#endif /* SUCCINCT_OUTPUT_DDJ_0504 */

	for(j = 0; j < WARMUPS; ++j)
	{
		counter.start();
		for(i = 0, iRet = 0; i < iterations; ++i)
		{
			// (i) Allocate
			buffer_t	buffer(alloc_size);

			// (ii) Use
//			s_p = const_cast<volatile element_t**>(&buffer);

			// Help to avoid compiler optimising away our code, by touching 
			// last element in the array ...
			buffer[alloc_size - 1] += static_cast<element_t>(i);
			// ... and testing the next to last one
			iRet ^= buffer[alloc_size - 1];

			// (iii) Deallocate
			; // Destructor will handle
		}
		counter.stop();
	}

#ifdef SUCCINCT_OUTPUT_DDJ_0504
	fprintf(stdout, test_printf_fmt " in %d repeats\n", counter.get_milliseconds(), iterations);
#else /* ? _DEBUG */
	fprintf(stdout, " \t" test_printf_fmt "\n", counter.get_milliseconds());
#endif /* _DEBUG */

	return iRet;
}

/* ////////////////////////////////////////////////////////////////////////// */
