/* /////////////////////////////////////////////////////////////////////////////
 * File:		rectarr.cpp
 *
 * Purpose:		Test program for Rectangular arrays
 *
 * Created		16th June 2003
 * Updated:		21st February 2004
 *
 * Copyright:	Synesis Software Pty Ltd, (c) 2003. All rights reserved.
 *
 * Home:		www.synesis.com.au/software
 *
 * ////////////////////////////////////////////////////////////////////////// */

#define	NOMINMAX

#include <stdio.h>

typedef	double												element_t;

#include <winstl_processheap_allocator.h>

typedef winstl_ns_qual(processheap_allocator)<element_t>	ator_t;

#include <stlsoft_fixed_array.h>

typedef stlsoft_ns_qual(fixed_array_3d)<element_t, ator_t>	array_3d_t;


#if defined(WIN32) || \
    defined(_WIN32)

#include <winstl_performance_counter.h>
typedef winstl_ns_qual(performance_counter)		performance_counter_t;

#else

#include <unixstl_performance_counter.h>
typedef unixstl_ns_qual(performance_counter)	performance_counter_t;

#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * printf() format string
 */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(_INTEL_COMPILER) || \
    defined(_MSC_VER)
 #define _sint64_fmt    "%I64d"
 #define _uint64_fmt    "%I64u"
#elif defined(__DMC__) || \
      defined(__MWERKS__)
 #define _sint64_fmt    "%lld"
 #define _uint64_fmt    "%llu"
#else
 #error Further compiler discrimination is required
#endif /* compiler */

/* ////////////////////////////////////////////////////////////////////////// */

const int	WARMUPS			=	2;
const int	DEF_ITERATIONS	=	1;
const int	DEF_D0			=	10;
const int	DEF_D1			=	50;
const int	DEF_D2			=	100;

/* ////////////////////////////////////////////////////////////////////////// */

/* Avoid the painful mess that is compiling with max & Windows headers */
inline int int_max(int i1, int i2)
{
	return i1 < i2 ? i2 : i1;
}

/* ////////////////////////////////////////////////////////////////////////// */


int main(int argc, char **argv)
{
	performance_counter_t	counter;
	int						ITERATIONS	=	(argc > 1) ? atoi(argv[1]) : 0;
	int						D0			=	(argc > 2) ? atoi(argv[2]) : 0;
	int						D1			=	(argc > 3) ? atoi(argv[3]) : 0;
	int						D2			=	(argc > 4) ? atoi(argv[4]) : 0;
	int						w;
	double					total		=	0.0;

	ITERATIONS	=	int_max(ITERATIONS, DEF_ITERATIONS);
	D0			=	int_max(D0, DEF_D0);
	D1			=	int_max(D1, DEF_D1);
	D2			=	int_max(D2, DEF_D2);

#ifdef WIN32
	::SetPriorityClass(::GetCurrentProcess(), HIGH_PRIORITY_CLASS);
	::SetThreadPriority(::GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
#endif // WIN32

	for(w = 0; w < WARMUPS; ++w)
	{
		int		i;

		counter.start();
		{
			array_3d_t	array(D0, D1, D2);

			for(i = 0; i < ITERATIONS; ++i)
			{
				// (i) Create the array
				int	i0;
				int	i1;
				int	i2;

				for(i0 = 0; i0 < D0; ++i0)
				{
					for(i1 = 0; i1 < D1; ++i1)
					{
						for(i2 = 0; i2 < D2; ++i2)
						{
							// Use unchecked efficient access here ...
							array.at_unchecked(i0, i1, i2) = static_cast<double>(::GetVersion()) * i0 * i1 * i2;

							// ... and checked & indexed access here.
							total += array[i0][i1][i2];
						}
					}
				}
			}
		}
		counter.stop();
	}

	fprintf(stdout,
			"fixed_array (%s):\t" _sint64_fmt " in %d repeats;\t[%d, %d, %d]: %.2G\n",
			__STLSOFT_COMPILER_LABEL_STRING,
			counter.get_milliseconds(),
			ITERATIONS,
			D0,
			D1,
			D2,
			total);

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
