/* ////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_64bit_integers.h (originally MWLrgInt.h, ::SynesisWin)
 *
 * Purpose:     Byte-order independent large integer (aggregate) classes.
 *
 * Created:     15th October 2000
 * Updated:     23rd November 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _STLSOFT_INCL_H_STLSOFT_64BIT_INTEGERS
#define _STLSOFT_INCL_H_STLSOFT_64BIT_INTEGERS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define _STLSOFT_VER_H_STLSOFT_64BIT_INTEGERS_MAJOR    2
# define _STLSOFT_VER_H_STLSOFT_64BIT_INTEGERS_MINOR    1
# define _STLSOFT_VER_H_STLSOFT_64BIT_INTEGERS_REVISION 1
# define _STLSOFT_VER_H_STLSOFT_64BIT_INTEGERS_EDIT     93
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

class sinteger64
{
public:
    typedef sinteger64  class_type;

public:
    sinteger64();
    sinteger64(ss_sint8_t i);
    sinteger64(ss_sint16_t i);
    sinteger64(ss_sint32_t i);
    sinteger64(ss_sint32_t high, ss_sint32_t low);
    sinteger64(int i);

    class_type &operator =(class_type const &rhs);
    class_type &operator =(ss_sint8_t i);
    class_type &operator =(ss_sint16_t i);
    class_type &operator =(ss_sint32_t i);
    class_type &operator =(int i);

    class_type &operator +=(class_type const &rhs);
    class_type &operator +=(ss_sint8_t i);
    class_type &operator +=(ss_sint16_t i);
    class_type &operator +=(ss_sint32_t i);
    class_type &operator +=(int i);

    class_type &operator -=(class_type const &rhs);
    class_type &operator -=(ss_sint8_t i);
    class_type &operator -=(ss_sint16_t i);
    class_type &operator -=(ss_sint32_t i);
    class_type &operator -=(int i);

    class_type &operator *=(class_type const &rhs);
    class_type &operator *=(ss_sint8_t i);
    class_type &operator *=(ss_sint16_t i);
    class_type &operator *=(ss_sint32_t i);
    class_type &operator *=(int i);

    class_type &operator /=(class_type const &rhs);
    class_type &operator /=(ss_sint8_t i);
    class_type &operator /=(ss_sint16_t i);
    class_type &operator /=(ss_sint32_t i);
    class_type &operator /=(int i);

public:
    ss_sint_t compare(class_type const &rhs) const;
    ss_sint_t compare(ss_sint8_t i) const;
    ss_sint_t compare(ss_sint16_t i) const;
    ss_sint_t compare(ss_sint32_t i) const;
    ss_sint_t compare(int i) const;

// Members
private:
    ss_sint32_t m_low;
    ss_sint32_t m_high;
};

/* /////////////////////////////////////////////////////////////////////////////
 * Operators
 */

// operator ==

template <ss_typename_param_k T>
inline ss_bool_t operator ==(sinteger64 const &lhs, T const &rhs)
{
    return 0 == lhs.compare(rhs);
}

template <ss_typename_param_k T>
inline ss_bool_t operator ==(T const &lhs, sinteger64 const &rhs)
{
    return 0 == rhs.compare(lhs);
}

// operator !=

template <ss_typename_param_k T>
inline ss_bool_t operator !=(sinteger64 const &lhs, T const &rhs)
{
    return 0 != lhs.compare(rhs);
}

template <ss_typename_param_k T>
inline ss_bool_t operator !=(T const &lhs, sinteger64 const &rhs)
{
    return 0 != rhs.compare(lhs);
}


// operator <

template <ss_typename_param_k T>
inline ss_bool_t operator <(sinteger64 const &lhs, T const &rhs)
{
    return lhs.compare(rhs) < 0;
}


// operator +

template <ss_typename_param_k T>
inline sinteger64 operator +(sinteger64 const &lhs, T const &rhs)
{
    sinteger64  result(lhs);

    result += rhs;

    return result;
}

// operator -

template <ss_typename_param_k T>
inline sinteger64 operator -(sinteger64 const &lhs, T const &rhs)
{
    sinteger64  result(lhs);

    result -= rhs;

    return result;
}

// operator *

template <ss_typename_param_k T>
inline sinteger64 operator *(sinteger64 const &lhs, T const &rhs)
{
    sinteger64  result(lhs);

    result *= rhs;

    return result;
}

// operator /

template <ss_typename_param_k T>
inline sinteger64 operator /(sinteger64 const &lhs, T const &rhs)
{
    sinteger64  result(lhs);

    result /= rhs;

    return result;
}


/* /////////////////////////////////////////////////////////////////////////////
 * Implementation
 */

inline sinteger64::sinteger64()
{}

inline sinteger64::sinteger64(ss_sint8_t i)
    : m_high(0)
    , m_low(i)
{}

inline sinteger64::sinteger64(ss_sint16_t i)
    : m_high(0)
    , m_low(i)
{}

inline sinteger64::sinteger64(ss_sint32_t i)
    : m_high(0)
    , m_low(i)
{}

inline sinteger64::sinteger64(ss_sint32_t high, ss_sint32_t low)
    : m_high(high)
    , m_low(low)
{}

inline sinteger64::sinteger64(int i)
{}

inline sinteger64::class_type &sinteger64::operator =(class_type const &rhs)
{
    m_high  =   rhs.m_high;
    m_low   =   rhs.m_low;

    return *this;
}

inline sinteger64::class_type &sinteger64::operator =(ss_sint8_t i)
{
    m_high  =   0;
    m_low   =   i;

    return *this;
}

inline sinteger64::class_type &sinteger64::operator =(ss_sint16_t i)
{
    m_high  =   0;
    m_low   =   i;

    return *this;
}

inline sinteger64::class_type &sinteger64::operator =(ss_sint32_t i)
{
    m_high  =   0;
    m_low   =   i;

    return *this;
}

inline sinteger64::class_type &sinteger64::operator =(int i)
{
    m_high  =   0;
    m_low   =   i;

    return *this;
}

inline sinteger64::class_type &sinteger64::operator +=(class_type const &rhs)
{
    m_low   +=  rhs.m_low;
    m_high  +=  rhs.m_high;

    if( rhs.m_low > 0 && 
        m_low < rhs.m_low)
    {
        ++m_high;
    }
    else if(rhs.m_low < 0 && 
            m_low > rhs.m_low)
    {
        --m_high;
    }

    return *this;
}

inline sinteger64::class_type &sinteger64::operator +=(ss_sint8_t i)
{
    return operator +=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator +=(ss_sint16_t i)
{
    return operator +=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator +=(ss_sint32_t i)
{
    return operator +=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator +=(int i)
{
    return operator +=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator -=(class_type const &rhs)
{
    m_low   -=  rhs.m_low;
    m_high  -=  rhs.m_high;

    if( rhs.m_low > 0 && 
        m_low < rhs.m_low)
    {
        ++m_high;
    }
    else if(rhs.m_low < 0 && 
            m_low > rhs.m_low)
    {
        --m_high;
    }

    return *this;
}

inline sinteger64::class_type &sinteger64::operator -=(ss_sint8_t i)
{
    return operator -=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator -=(ss_sint16_t i)
{
    return operator -=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator -=(ss_sint32_t i)
{
    return operator -=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator -=(int i)
{
    return operator -=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator *=(class_type const &rhs)
{
    m_low   *=  rhs.m_low;
    m_high  *=  rhs.m_high;

    if( rhs.m_low > 0 && 
        m_low < rhs.m_low)
    {
        ++m_high;
    }
    else if(rhs.m_low < 0 && 
            m_low > rhs.m_low)
    {
        --m_high;
    }

    return *this;
}

inline sinteger64::class_type &sinteger64::operator *=(ss_sint8_t i)
{
    return operator *=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator *=(ss_sint16_t i)
{
    return operator *=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator *=(ss_sint32_t i)
{
    return operator *=(class_type(i));
}

inline sinteger64::class_type &sinteger64::operator *=(int i)
{
    return operator *=(class_type(i));
}

inline ss_sint_t sinteger64::compare(class_type const &rhs) const
{
    ss_sint_t   res =   m_high - rhs.m_high;

    if(0 == res)
    {
        res = m_low - rhs.m_low;
    }

    return res;
}

inline ss_sint_t sinteger64::compare(ss_sint8_t i) const
{
    return compare(class_type(0, i));
}

inline ss_sint_t sinteger64::compare(ss_sint16_t i) const
{
    return compare(class_type(0, i));
}

inline ss_sint_t sinteger64::compare(ss_sint32_t i) const
{
    return compare(class_type(0, i));
}

inline ss_sint_t sinteger64::compare(int i) const
{
    return compare(class_type(0, i));
}


/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_STLSOFT_INCL_H_STLSOFT_64BIT_INTEGERS */

/* ////////////////////////////////////////////////////////////////////////// */
