/* ////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_allocator_base.h
 *
 * Purpose:     Allocator commmon features.
 *
 * Created:     20th August 2003
 * Updated:     17th November 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _STLSOFT_INCL_H_STLSOFT_ALLOCATOR_BASE
#define _STLSOFT_INCL_H_STLSOFT_ALLOCATOR_BASE

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
#define _STLSOFT_VER_H_STLSOFT_ALLOCATOR_BASE_MAJOR     1
#define _STLSOFT_VER_H_STLSOFT_ALLOCATOR_BASE_MINOR     0
#define _STLSOFT_VER_H_STLSOFT_ALLOCATOR_BASE_REVISION  3
#define _STLSOFT_VER_H_STLSOFT_ALLOCATOR_BASE_EDIT      3
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */
#if defined(__STLSOFT_CF_THROW_BAD_ALLOC) || \
    (   defined(__STLSOFT_COMPILER_IS_MSVC) && \
        _MSC_VER < 1100)
# include <new>         // placement new, std::bad_alloc
#endif /* __STLSOFT_CF_THROW_BAD_ALLOC || _MSC_VER < 1100) */

/* ////////////////////////////////////////////////////////////////////////////
 * Allocator compatibilities
 */

/* stl-like allocator classes need allocate and deallocate as static
 *
 * Discriminated symbol is __STLSOFT_CF_ALLOCATOR_STATIC_ALLOCATE_METHODS
 *
 * Note: this should be resolving on the library, not the compiler
 */
#ifdef __STLSOFT_CF_ALLOCATOR_STATIC_ALLOCATE_METHODS
# undef __STLSOFT_CF_ALLOCATOR_STATIC_ALLOCATE_METHODS
#endif /* __STLSOFT_CF_ALLOCATOR_STATIC_ALLOCATE_METHODS */

/* stl-like allocator classes need deallocate to use void *
 *
 * Discriminated symbol is __STLSOFT_CF_ALLOCATOR_TYPED_DEALLOCATE_POINTER
 *
 * Note: this should be resolving on the library, not the compiler
 */
#define __STLSOFT_CF_ALLOCATOR_TYPED_DEALLOCATE_POINTER /* This is standard behaviour */

/* stl-like allocator classes need a _CharAlloc method
 *
 * Discriminated symbol is __STLSOFT_CF_ALLOCATOR_CHARALLOC_METHOD
 *
 * Note: this should be resolving on the library, not the compiler
 */
#ifdef __STLSOFT_CF_ALLOCATOR_CHARALLOC_METHOD
# undef __STLSOFT_CF_ALLOCATOR_CHARALLOC_METHOD
#endif /* __STLSOFT_CF_ALLOCATOR_CHARALLOC_METHOD */


#if defined(__STLSOFT_COMPILER_IS_DMC)
# define __STLSOFT_CF_ALLOCATOR_STATIC_ALLOCATE_METHODS
# undef __STLSOFT_CF_ALLOCATOR_TYPED_DEALLOCATE_POINTER
#elif defined(__STLSOFT_COMPILER_IS_MWERKS)
# undef __STLSOFT_CF_ALLOCATOR_TYPED_DEALLOCATE_POINTER
#elif ( defined(__STLSOFT_COMPILER_IS_INTEL) || \
        defined(__STLSOFT_COMPILER_IS_MSVC)) && \
    _MSC_VER < 1300
# undef __STLSOFT_CF_ALLOCATOR_TYPED_DEALLOCATE_POINTER
# define __STLSOFT_CF_ALLOCATOR_CHARALLOC_METHOD
#endif /* compiler */

#if defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
# ifdef __STLSOFT_CF_ALLOCATOR_CHARALLOC_METHOD
#  undef __STLSOFT_CF_ALLOCATOR_CHARALLOC_METHOD
# endif /* __STLSOFT_CF_ALLOCATOR_CHARALLOC_METHOD */
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */


/* ////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \weakgroup libraries STLSoft Libraries
/// \brief The individual libraries

/// \weakgroup libraries_allocator Allocator Library
/// \ingroup libraries
/// \brief This library provides STL-compatible <code><b>allocator</b></code> types

/// \weakgroup stlsoft_allocator_library Allocator Library (STLSoft)
/// \ingroup STLSoft libraries_allocator
/// \brief This library provides STL-compatible <code><b>allocator</b></code> types
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */


/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group stlsoft_allocator_library

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_STLSOFT_INCL_H_STLSOFT_ALLOCATOR_BASE */

/* ////////////////////////////////////////////////////////////////////////// */
