/* ////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_container_base.h
 *
 * Purpose:     Allocator commmon features.
 *
 * Created:     17th February 2004
 * Updated:     17th February 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _STLSOFT_INCL_H_STLSOFT_CONTAINER_BASE
#define _STLSOFT_INCL_H_STLSOFT_CONTAINER_BASE

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define _STLSOFT_VER_H_STLSOFT_CONTAINER_BASE_MAJOR    1
# define _STLSOFT_VER_H_STLSOFT_CONTAINER_BASE_MINOR    0
# define _STLSOFT_VER_H_STLSOFT_CONTAINER_BASE_REVISION 1
# define _STLSOFT_VER_H_STLSOFT_CONTAINER_BASE_EDIT     1
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */

/* ////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////////
 * Classes
 */

template <ss_typename_param_k A>
struct container_base
#ifndef STLSOFT_CF_ALLOCATOR_BASE_EXPENSIVE
    : public A
#endif /* STLSOFT_CF_ALLOCATOR_BASE_EXPENSIVE */
{
    typedef A                   allocator_type;
    typedef container_base<A>   class_type;

protected:
    container_base()
    {}
    container_base(class_type const &)
    {}

	operator allocator_type &()
	{
		return get_allocator_();
	}
	operator allocator_type const &() const
	{
		return get_allocator_();
	}

private:
#ifndef STLSOFT_CF_ALLOCATOR_BASE_EXPENSIVE
    allocator_type &get_allocator_()
    {
        return *this;
    }
    allocator_type const &get_allocator_() const
    {
        return *this;
    }
#else /* ? STLSOFT_CF_ALLOCATOR_BASE_EXPENSIVE */
    static allocator_type &get_allocator_()
    {
        static allocator_type	s_allocator;

        return s_allocator;
    }
#endif /* STLSOFT_CF_ALLOCATOR_BASE_EXPENSIVE */
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_STLSOFT_INCL_H_STLSOFT_CONTAINER_BASE */

/* ////////////////////////////////////////////////////////////////////////// */
