/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_limit_traits.h
 *
 * Purpose:     limit_traits classes. Provides nothing that is not in
 *              std::numeric_limits, but uses minimum() and maximum() rather
 *              than min() and max(), since some compilers are not well-behaved
 *              in making these functions rather than macros.
 *
 * Created:     16th January 2002
 * Updated:     15th January 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef STLSOFT_INCL_H_STLSOFT_LIMIT_TRAITS
#define STLSOFT_INCL_H_STLSOFT_LIMIT_TRAITS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define _STLSOFT_VER_H_STLSOFT_LIMIT_TRAITS_MAJOR       1
# define _STLSOFT_VER_H_STLSOFT_LIMIT_TRAITS_MINOR       5
# define _STLSOFT_VER_H_STLSOFT_LIMIT_TRAITS_REVISION    3
# define _STLSOFT_VER_H_STLSOFT_LIMIT_TRAITS_EDIT        26
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants & definitions
 */

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

#define __STLSOFT_GEN_SINT8_SUFFIX(i)           (i)
#define __STLSOFT_GEN_UINT8_SUFFIX(i)           (i ## U)
#define __STLSOFT_GEN_SINT16_SUFFIX(i)          (i)
#define __STLSOFT_GEN_UINT16_SUFFIX(i)          (i ## U)
#define __STLSOFT_GEN_SINT32_SUFFIX(i)          (i ## L)
#define __STLSOFT_GEN_UINT32_SUFFIX(i)          (i ## UL)
#if (   defined(__STLSOFT_COMPILER_IS_DMC) || \
        defined(__STLSOFT_COMPILER_IS_COMO) || \
        defined(__STLSOFT_COMPILER_IS_GCC) || \
        defined(__STLSOFT_COMPILER_IS_MWERKS))
# define __STLSOFT_GEN_SINT64_SUFFIX(i)         (i ## LL)
# define __STLSOFT_GEN_UINT64_SUFFIX(i)         (i ## ULL)
#elif ( defined(__STLSOFT_COMPILER_IS_BORLAND) || \
        defined(__STLSOFT_COMPILER_IS_INTEL) || \
        defined(__STLSOFT_COMPILER_IS_MSVC) || \
        defined(__STLSOFT_COMPILER_IS_VECTORC) || \
        defined(__STLSOFT_COMPILER_IS_WATCOM))
# define __STLSOFT_GEN_SINT64_SUFFIX(i)         (i ## L)
# define __STLSOFT_GEN_UINT64_SUFFIX(i)         (i ## UL)
#else
#error Compiler not discriminated
#endif /* __STLSOFT_DVS_COMPILER_GNUC || __STLSOFT_DVS_COMPILER_DECCXX */


#define __STLSOFT_LIMIT_TRAITS__SINT8_MIN       (- __STLSOFT_GEN_SINT8_SUFFIX(127) - 1)
#define __STLSOFT_LIMIT_TRAITS__SINT8_MAX       (+ __STLSOFT_GEN_SINT8_SUFFIX(127))

#define __STLSOFT_LIMIT_TRAITS__UINT8_MIN       (  __STLSOFT_GEN_UINT8_SUFFIX(0))
#define __STLSOFT_LIMIT_TRAITS__UINT8_MAX       (  __STLSOFT_GEN_UINT8_SUFFIX(255))


#define __STLSOFT_LIMIT_TRAITS__SINT16_MIN      (- __STLSOFT_GEN_SINT16_SUFFIX(32767) - 1)
#define __STLSOFT_LIMIT_TRAITS__SINT16_MAX      (+ __STLSOFT_GEN_SINT16_SUFFIX(32767))

#define __STLSOFT_LIMIT_TRAITS__UINT16_MIN      (  __STLSOFT_GEN_UINT16_SUFFIX(0))
#define __STLSOFT_LIMIT_TRAITS__UINT16_MAX      (  __STLSOFT_GEN_UINT16_SUFFIX(65535))


#define __STLSOFT_LIMIT_TRAITS__SINT32_MIN      (- __STLSOFT_GEN_SINT32_SUFFIX(2147483647) - 1)
#define __STLSOFT_LIMIT_TRAITS__SINT32_MAX      (+ __STLSOFT_GEN_SINT32_SUFFIX(2147483647))

#define __STLSOFT_LIMIT_TRAITS__UINT32_MIN      (  __STLSOFT_GEN_UINT32_SUFFIX(0))
#define __STLSOFT_LIMIT_TRAITS__UINT32_MAX      (  __STLSOFT_GEN_UINT32_SUFFIX(4294967295))


#define __STLSOFT_LIMIT_TRAITS__SINT64_MIN      (- __STLSOFT_GEN_SINT64_SUFFIX(9223372036854775807) - 1)
#define __STLSOFT_LIMIT_TRAITS__SINT64_MAX      (+ __STLSOFT_GEN_SINT64_SUFFIX(9223372036854775807) )

#define __STLSOFT_LIMIT_TRAITS__UINT64_MIN      (  __STLSOFT_GEN_UINT64_SUFFIX(0) )
#define __STLSOFT_LIMIT_TRAITS__UINT64_MAX      (  __STLSOFT_GEN_UINT64_SUFFIX(18446744073709551615) )

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

#ifdef __STLSOFT_DOCUMENTATION_SKIP_SECTION
/// Traits for determining the minimum and maximum values of types
///
/// limit_traits is a traits class for acquiring the minimum and maximum values
/// of types.
///
/// &nbsp;&nbsp;<code>limit_traits<ss_sint16_t>::minimum() == -32768</code><br>
/// &nbsp;&nbsp;<code>limit_traits<ss_sint16_t>::maximum() == 32767</code><br>
///
/// \param T The type
template <ss_typename_param_k T>
struct limit_traits
{
public:
    /// The value type
    typedef T value_type;

public:
    /// Returns the minimum value for the type
    static value_type       minimum();
    /// Returns the maximum value for the type
    static value_type       maximum();

#ifdef STLSOFT_CF_MEMBER_CONSTANT_SUPPORT
    static const value_type minimum_value = ???;
    static const value_type maximum_value = ???;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};

#else

// class limit_traits

template <ss_typename_param_k T>
struct limit_traits;

/* s/uint8 */
STLSOFT_TEMPLATE_SPECIALISATION
struct limit_traits<ss_sint8_t>
{
public:
    typedef ss_sint8_t  value_type;

public:
    static value_type       minimum() { return __STLSOFT_LIMIT_TRAITS__SINT8_MIN; }
    static value_type       maximum() { return __STLSOFT_LIMIT_TRAITS__SINT8_MAX; }

#ifdef STLSOFT_CF_MEMBER_CONSTANT_SUPPORT
    static const value_type minimum_value = __STLSOFT_LIMIT_TRAITS__SINT8_MIN;
    static const value_type maximum_value = __STLSOFT_LIMIT_TRAITS__SINT8_MAX;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};

STLSOFT_TEMPLATE_SPECIALISATION
struct limit_traits<ss_uint8_t>
{
public:
    typedef ss_uint8_t  value_type;

public:
    static value_type   minimum() { return __STLSOFT_LIMIT_TRAITS__UINT8_MIN; }
    static value_type   maximum() { return __STLSOFT_LIMIT_TRAITS__UINT8_MAX; }

#ifdef STLSOFT_CF_MEMBER_CONSTANT_SUPPORT
    static const value_type minimum_value = __STLSOFT_LIMIT_TRAITS__UINT8_MIN;
    static const value_type maximum_value = __STLSOFT_LIMIT_TRAITS__UINT8_MAX;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};

/* s/uint16 */
STLSOFT_TEMPLATE_SPECIALISATION
struct limit_traits<ss_sint16_t>
{
public:
    typedef ss_sint16_t value_type;

public:
    static value_type   minimum() { return __STLSOFT_LIMIT_TRAITS__SINT16_MIN; }
    static value_type   maximum() { return __STLSOFT_LIMIT_TRAITS__SINT16_MAX; }

#ifdef STLSOFT_CF_MEMBER_CONSTANT_SUPPORT
    static const value_type minimum_value = __STLSOFT_LIMIT_TRAITS__SINT16_MIN;
    static const value_type maximum_value = __STLSOFT_LIMIT_TRAITS__SINT16_MAX;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};

STLSOFT_TEMPLATE_SPECIALISATION
struct limit_traits<ss_uint16_t>
{
public:
    typedef ss_uint16_t value_type;

public:
    static value_type   minimum() { return __STLSOFT_LIMIT_TRAITS__UINT16_MIN; }
    static value_type   maximum() { return __STLSOFT_LIMIT_TRAITS__UINT16_MAX; }

#ifdef STLSOFT_CF_MEMBER_CONSTANT_SUPPORT
    static const value_type minimum_value = __STLSOFT_LIMIT_TRAITS__UINT16_MIN;
    static const value_type maximum_value = __STLSOFT_LIMIT_TRAITS__UINT16_MAX;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};

/* s/uint32 */
STLSOFT_TEMPLATE_SPECIALISATION
struct limit_traits<ss_sint32_t>
{
public:
    typedef ss_sint32_t value_type;

public:
    static value_type   minimum() { return __STLSOFT_LIMIT_TRAITS__SINT32_MIN; }
    static value_type   maximum() { return __STLSOFT_LIMIT_TRAITS__SINT32_MAX; }

#ifdef STLSOFT_CF_MEMBER_CONSTANT_SUPPORT
    static const value_type minimum_value = __STLSOFT_LIMIT_TRAITS__SINT32_MIN;
    static const value_type maximum_value = __STLSOFT_LIMIT_TRAITS__SINT32_MAX;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};

STLSOFT_TEMPLATE_SPECIALISATION
struct limit_traits<ss_uint32_t>
{
public:
    typedef ss_uint32_t value_type;

public:
    static value_type   minimum() { return __STLSOFT_LIMIT_TRAITS__UINT32_MIN; }
    static value_type   maximum() { return __STLSOFT_LIMIT_TRAITS__UINT32_MAX; }

#ifdef STLSOFT_CF_MEMBER_CONSTANT_SUPPORT
    static const value_type minimum_value = __STLSOFT_LIMIT_TRAITS__UINT32_MIN;
    static const value_type maximum_value = __STLSOFT_LIMIT_TRAITS__UINT32_MAX;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};

#ifdef STLSOFT_CF_64BIT_INT_SUPPORT

/* s/uint64 */
STLSOFT_TEMPLATE_SPECIALISATION
struct limit_traits<ss_sint64_t>
{
public:
    typedef ss_sint64_t value_type;

public:
    static value_type   minimum() { return __STLSOFT_LIMIT_TRAITS__SINT64_MIN; }
    static value_type   maximum() { return __STLSOFT_LIMIT_TRAITS__SINT64_MAX; }

#if defined(STLSOFT_CF_MEMBER_CONSTANT_SUPPORT) && \
    !defined(__STLSOFT_COMPILER_IS_BORLAND)
    static const value_type minimum_value = __STLSOFT_LIMIT_TRAITS__SINT64_MIN;
    static const value_type maximum_value = __STLSOFT_LIMIT_TRAITS__SINT64_MAX;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};

STLSOFT_TEMPLATE_SPECIALISATION
struct limit_traits<ss_uint64_t>
{
public:
    typedef ss_uint64_t value_type;

public:
    static value_type   minimum() { return __STLSOFT_LIMIT_TRAITS__UINT64_MIN; }
    static value_type   maximum() { return __STLSOFT_LIMIT_TRAITS__UINT64_MAX; }

#if defined(STLSOFT_CF_MEMBER_CONSTANT_SUPPORT) && \
    !defined(__STLSOFT_COMPILER_IS_BORLAND)
    static const value_type minimum_value = __STLSOFT_LIMIT_TRAITS__UINT64_MIN;
    static const value_type maximum_value = __STLSOFT_LIMIT_TRAITS__UINT64_MAX;
#endif /* STLSOFT_CF_MEMBER_CONSTANT_SUPPORT */
};
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_LIMIT_TRAITS */

/* ////////////////////////////////////////////////////////////////////////// */
