/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_printf_traits.h
 *
 * Purpose:     printf_traits classes.
 *
 * Created:     16th January 2002
 * Updated:     15th January 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef STLSOFT_INCL_H_STLSOFT_PRINTF_TRAITS
#define STLSOFT_INCL_H_STLSOFT_PRINTF_TRAITS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define _STLSOFT_VER_H_STLSOFT_PRINTF_TRAITS_MAJOR       1
# define _STLSOFT_VER_H_STLSOFT_PRINTF_TRAITS_MINOR       1
# define _STLSOFT_VER_H_STLSOFT_PRINTF_TRAITS_REVISION    4
# define _STLSOFT_VER_H_STLSOFT_PRINTF_TRAITS_EDIT        16
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
 #include "stlsoft.h"   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants & definitions
 */

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

#define __STLSOFT_PRINTF_TRAITS__SINT8_MIN      -128
#define __STLSOFT_PRINTF_TRAITS__SINT8_MAX      127

#define __STLSOFT_PRINTF_TRAITS__UINT8_MIN      0
#define __STLSOFT_PRINTF_TRAITS__UINT8_MAX      255

#define __STLSOFT_PRINTF_TRAITS__SINT16_MIN     -32768
#define __STLSOFT_PRINTF_TRAITS__SINT16_MAX     32767

#define __STLSOFT_PRINTF_TRAITS__UINT16_MIN     0
#define __STLSOFT_PRINTF_TRAITS__UINT16_MAX     65535

#define __STLSOFT_PRINTF_TRAITS__SINT32_MIN     -2147483648
#define __STLSOFT_PRINTF_TRAITS__SINT32_MAX     2147483647

#define __STLSOFT_PRINTF_TRAITS__UINT32_MIN     0
#define __STLSOFT_PRINTF_TRAITS__UINT32_MAX     4294967295

#define __STLSOFT_PRINTF_TRAITS__SINT64_MIN     -9223372036854775808
#define __STLSOFT_PRINTF_TRAITS__SINT64_MAX     9223372036854775807

#define __STLSOFT_PRINTF_TRAITS__UINT64_MIN     0
#define __STLSOFT_PRINTF_TRAITS__UINT64_MAX     18446744073709551615

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Macros
 */

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

#define ___stringise_(x)            # x
#define ___stringise(x)             ___stringise_(x)

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// struct printf_traits

#ifdef __STLSOFT_DOCUMENTATION_SKIP_SECTION
/// Traits for determining the size, in printf-ed form, of the minimum and 
/// maximum values of types
///
/// printf_traits is a traits class for acquiring enum values representing the
/// lengths, when expressed in string form, of the minimum and maximum values of
/// the type, and the maximum of the two. The lengths are inclusive of the
/// null terminator.
///
/// &nbsp;&nbsp;<code>printf_traits<ss_sint16_t>::size_min == 6</code><br>
/// &nbsp;&nbsp;<code>printf_traits<ss_sint16_t>::size_max == 7</code><br>
/// &nbsp;&nbsp;<code>printf_traits<ss_sint16_t>::size == 7</code><br>
///
/// \param T The type
template <ss_typename_param_k T>
struct printf_traits
{
    enum
    {
            size_min    //!< The number of characters (& null) in the minimum value
        ,   size_max    //!< The number of characters (& null) in the maximum value
        ,   size        //!< The maximum of \c size_min and \c size_max
    };
};

#else

template <ss_typename_param_k T>
struct printf_traits;

/* s/uint8 */
#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
template <>
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */
struct printf_traits<ss_sint8_t>
{
    enum
    {
            size_min = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__SINT8_MIN))
        ,   size_max = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__SINT8_MAX))
        ,   size = (size_min < size_max) ? size_max : size_min
    };
};

#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
template <>
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */
struct printf_traits<ss_uint8_t>
{
    enum
    {
            size_min = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__UINT8_MIN))
        ,   size_max = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__UINT8_MAX))
        ,   size = (size_min < size_max) ? size_max : size_min
    };
};

/* s/uint16 */
#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
template <>
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */
struct printf_traits<ss_sint16_t>
{
    enum
    {
            size_min = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__SINT16_MIN))
        ,   size_max = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__SINT16_MAX))
        ,   size = (size_min < size_max) ? size_max : size_min
    };
};

#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
template <>
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */
struct printf_traits<ss_uint16_t>
{
    enum
    {
            size_min = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__UINT16_MIN))
        ,   size_max = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__UINT16_MAX))
        ,   size = (size_min < size_max) ? size_max : size_min
    };
};

/* s/uint32 */
#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
template <>
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */
struct printf_traits<ss_sint32_t>
{
    enum
    {
            size_min = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__SINT32_MIN))
        ,   size_max = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__SINT32_MAX))
        ,   size = (size_min < size_max) ? size_max : size_min
    };
};

#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
template <>
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */
struct printf_traits<ss_uint32_t>
{
    enum
    {
            size_min = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__UINT32_MIN))
        ,   size_max = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__UINT32_MAX))
        ,   size = (size_min < size_max) ? size_max : size_min
    };
};

#ifdef STLSOFT_CF_64BIT_INT_SUPPORT

/* s/uint64 */
#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
template <>
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */
struct printf_traits<ss_sint64_t>
{
    enum
    {
            size_min = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__SINT64_MIN))
        ,   size_max = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__SINT64_MAX))
        ,   size = (size_min < size_max) ? size_max : size_min
    };
};

#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
template <>
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */
struct printf_traits<ss_uint64_t>
{
    enum
    {
            size_min = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__UINT64_MIN))
        ,   size_max = sizeof(___stringise(__STLSOFT_PRINTF_TRAITS__UINT64_MAX))
        ,   size = (size_min < size_max) ? size_max : size_min
    };
};

#endif /* !STLSOFT_CF_64BIT_INT_SUPPORT */

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_PRINTF_TRAITS */

/* ////////////////////////////////////////////////////////////////////////// */
