/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_simple_type_converter.h (originally MLTypCnv.h, ::SynesisStd)
 *
 * Purpose:     union_cast - a dangerous weapon in the hands of the unwary.
 *
 * Created:     2nd May 1997
 * Updated:     28th December 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _STLSOFT_INCL_H_STLSOFT_SIMPLE_TYPE_CONVERTER
#define _STLSOFT_INCL_H_STLSOFT_SIMPLE_TYPE_CONVERTER

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define _STLSOFT_VER_H_STLSOFT_SIMPLE_TYPE_CONVERTER_MAJOR     2
# define _STLSOFT_VER_H_STLSOFT_SIMPLE_TYPE_CONVERTER_MINOR     1
# define _STLSOFT_VER_H_STLSOFT_SIMPLE_TYPE_CONVERTER_REVISION  2
# define _STLSOFT_VER_H_STLSOFT_SIMPLE_TYPE_CONVERTER_EDIT      10
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"               // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */
#ifndef _STLSOFT_INCL_H_STLSOFT_UNION_CAST
# include "stlsoft_union_cast.h"    // union_cast
#endif /* !_STLSOFT_INCL_H_STLSOFT_UNION_CAST */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// 

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// This class (union) effects conversion from one type to another, without
/// the use of casts.
///
/// \note The use of this technique is non-portable, and you use this class at
/// your own risk. Notwithstanding that, the TO and FROM types have to be the
/// same size, so the technique is widely usable.
template<   ss_typename_param_k TO
        ,   ss_typename_param_k FROM
        >
class simple_type_conversion
{
private:
    typedef union_cast<TO, FROM>                cast_type;
    typedef TO                                  to_type;
    typedef FROM                                from_type;
    typedef simple_type_conversion<TO, FROM>    class_type;

// Construction
public:
    /// Conversion constructor
    ss_explicit_k simple_type_conversion(from_type const from)
        : parent_class_type(from)
    {}

private:
    cast_type   m_cast;

// Not to be implemented
private:
    class_type &operator =(class_type const &);
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_STLSOFT_INCL_H_STLSOFT_SIMPLE_TYPE_CONVERTER */

/* ////////////////////////////////////////////////////////////////////////// */
