/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_type_traits.h
 *
 * Purpose:     type traits classes.
 *
 * Created:     24th August 2003
 * Updated:     21st December 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _STLSOFT_INCL_H_STLSOFT_TYPE_TRAITS
#define _STLSOFT_INCL_H_STLSOFT_TYPE_TRAITS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define _STLSOFT_VER_H_STLSOFT_TYPE_TRAITS_MAJOR       1
# define _STLSOFT_VER_H_STLSOFT_TYPE_TRAITS_MINOR       2
# define _STLSOFT_VER_H_STLSOFT_TYPE_TRAITS_REVISION    2
# define _STLSOFT_VER_H_STLSOFT_TYPE_TRAITS_EDIT        6
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

#ifdef __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT
/// \brief Deduces the base type 
template <typename T>
struct base_type_traits
{
    enum { is_pointer       =   0 };
    enum { is_reference     =   0 };
    enum { is_const         =   0 };
    enum { is_volatile      =   0 };

    typedef T   base_type;
    typedef T   cv_type;
};

template <typename T>
struct base_type_traits<T*>
{
    enum { is_pointer       =   1 };
    enum { is_reference     =   0 };
    enum { is_const         =   0 };
    enum { is_volatile      =   0 };

    typedef T   base_type;
    typedef T   cv_type;
};

template <typename T>
struct base_type_traits<T const*>
{
    enum { is_pointer       =   1 };
    enum { is_reference     =   0 };
    enum { is_const         =   1 };
    enum { is_volatile      =   0 };

    typedef T       base_type;
    typedef T const cv_type;
};

template <typename T>
struct base_type_traits<T volatile*>
{
    enum { is_pointer       =   1 };
    enum { is_reference     =   0 };
    enum { is_const         =   0 };
    enum { is_volatile      =   1 };

    typedef T           base_type;
    typedef T volatile  cv_type;
};

template <typename T>
struct base_type_traits<T const volatile*>
{
    enum { is_pointer       =   1 };
    enum { is_reference     =   0 };
    enum { is_const         =   1 };
    enum { is_volatile      =   1 };

    typedef T                   base_type;
    typedef T const volatile    cv_type;
};

template <typename T>
struct base_type_traits<T&>
{
    enum { is_pointer       =   0 };
    enum { is_reference     =   1 };
    enum { is_const         =   0 };
    enum { is_volatile      =   0 };

    typedef T   base_type;
    typedef T   cv_type;
};

template <typename T>
struct base_type_traits<T const&>
{
    enum { is_pointer       =   0 };
    enum { is_reference     =   1 };
    enum { is_const         =   1 };
    enum { is_volatile      =   0 };

    typedef T       base_type;
    typedef T const cv_type;
};

template <typename T>
struct base_type_traits<T volatile&>
{
    enum { is_pointer       =   0 };
    enum { is_reference     =   1 };
    enum { is_const         =   0 };
    enum { is_volatile      =   1 };

    typedef T           base_type;
    typedef T volatile  cv_type;
};

template <typename T>
struct base_type_traits<T const volatile&>
{
    enum { is_pointer       =   0 };
    enum { is_reference     =   1 };
    enum { is_const         =   1 };
    enum { is_volatile      =   1 };

    typedef T                   base_type;
    typedef T const volatile    cv_type;
};
#else /* ? __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT */

#endif // __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT



/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_STLSOFT_INCL_H_STLSOFT_TYPE_TRAITS */

/* ////////////////////////////////////////////////////////////////////////// */
