/* /////////////////////////////////////////////////////////////////////////////
 * File:		sprintf.cpp
 *
 * Purpose:		Test program for integer to string conversion
 *
 * Created		16th June 2003
 * Updated:		21st February 2004
 *
 * Copyright:	Synesis Software Pty Ltd, (c) 2003. All rights reserved.
 *
 * Home:		www.synesis.com.au/software
 *
 * ////////////////////////////////////////////////////////////////////////// */


#include <stdio.h>

#include <stlsoft.h>

stlsoft_ns_using(ss_uint32_t)

#if defined(INT2STRING_SPRINTF)
#elif defined(INT2STRING_STLSOFT)
 #include <stlsoft_integer_to_string.h>

 stlsoft_ns_using(integer_to_string)
#else
 #error Must specify one of INT2STRING_SPRINTF or INT2STRING_STLSOFT
#endif /* 0 */


#if defined(WIN32) || \
    defined(_WIN32)
 #include <winstl_performance_counter.h>
 
 typedef winstl_ns_qual(performance_counter)		performance_counter_t;
#else
 #include <unixstl_performance_counter.h>

 typedef unixstl_ns_qual(performance_counter)	performance_counter_t;
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * printf() format string
 */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(_INTEL_COMPILER) || \
    defined(_MSC_VER) || \
    defined(__WATCOMC__)
 #define _sint64_fmt    "%I64d"
 #define _uint64_fmt    "%I64u"
#elif defined(__DMC__) || \
      defined(__MWERKS__)
 #define _sint64_fmt    "%lld"
 #define _uint64_fmt    "%llu"
#else
 #error Further compiler discrimination is required
#endif /* compiler */

/* ////////////////////////////////////////////////////////////////////////// */

const int			WARMUPS		=	2;
const ss_uint32_t	DEF_I_MAX	=	100000;

/* ////////////////////////////////////////////////////////////////////////// */

/* Avoid the painful mess that is compiling with max & Windows headers */
inline int int_max(int i1, int i2)
{
	return i1 < i2 ? i2 : i1;
}

/* ////////////////////////////////////////////////////////////////////////// */

int main(int argc, char **argv)
{
	performance_counter_t	counter;
	int						w;
	ss_uint32_t				I_MAX	=	(argc > 1) ? atoi(argv[1]) : 0;
	int						len		=	0;

	I_MAX	=	int_max(I_MAX, DEF_I_MAX);

#ifdef WIN32
	::SetPriorityClass(::GetCurrentProcess(), HIGH_PRIORITY_CLASS);
	::SetThreadPriority(::GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
#endif // WIN32

	for(w = 0; w < WARMUPS; ++w)
	{
		char		szNum[21];
		ss_uint32_t	i;
#if defined(INT2STRING_SPRINTF)
#elif defined(INT2STRING_STLSOFT)
		char *const	end	=	szNum + stlsoft_num_elements(szNum);
#endif /* 0 */

		counter.start();

		for(i = 0, len = 0; i < I_MAX; ++i)
		{
#if defined(INT2STRING_SPRINTF)
			len += sprintf(szNum, "%u", i);
#elif defined(INT2STRING_STLSOFT)
			char const	*p	=	integer_to_string(szNum, stlsoft_num_elements(szNum), i);

			len += (end - p);
#endif /* 0 */
		}

		counter.stop();
	}

	fprintf(stdout,
#if defined(INT2STRING_SPRINTF)
			"int2string-sprintf (%s):\t" _sint64_fmt " in %d repeats; %d total\n",
#elif defined(INT2STRING_STLSOFT)
			"int2string-stlsoft (%s):\t" _sint64_fmt " in %d repeats; %d total\n",
#endif /* 0 */
			__STLSOFT_COMPILER_LABEL_STRING,
			counter.get_milliseconds(),
			(int)I_MAX,
			len);

	return len;
}

/* ////////////////////////////////////////////////////////////////////////// */
