/* /////////////////////////////////////////////////////////////////////////////
 * File:        pod_vector_speed_test.cpp
 *
 * Purpose:     Implementation file for the pod_vector_speed_test project.
 *
 * Created:     2nd January 2004
 * Updated:     21st February 2004
 *
 * Status:      Wizard-generated
 *
 * License:     (Licensed under the Synesis Software Open License)
 *
 *              Copyright (C) 1999-2004, Synesis Software Pty Ltd.
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/software
 *
 *              email:      software@synesis.com.au
 *
 *              This source code is placed into the public domain 2004
 *              by Synesis Software Pty Ltd. There are no restrictions
 *              whatsoever to your use of the software. 
 *
 *              This source code is provided by Synesis Software Pty Ltd "as is"
 *              and any warranties, whether expressed or implied, including, but
 *              not limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software Pty Ltd be liable for any direct,
 *              indirect, incidental, special, exemplary, or consequential
 *              damages (including, but not limited to, procurement of
 *              substitute goods or services; loss of use, data, or profits; or
 *              business interruption) however caused and on any theory of
 *              liability, whether in contract, strict liability, or tort
 *              (including negligence or otherwise) arising in any way out of
 *              the use of this software, even if advised of the possibility of
 *              such damage. 
 *
 *              Neither the name of Synesis Software Pty Ltd nor the names of
 *              any subdivisions, employees or agents of Synesis Software Pty
 *              Ltd, nor the names of any other contributors to this software
 *              may be used to endorse or promote products derived from this
 *              software without specific prior written permission. 
 *
 * ////////////////////////////////////////////////////////////////////////// */



#define SUCCINCT_OUTPUT_DDJ_0504



#include <windows.h>
#include <stdio.h>

#include <stlsoft.h>

#include <stlsoft_pod_vector.h>

stlsoft_ns_using(pod_vector)

#include <winstl_processheap_allocator.h>

winstl_ns_using(processheap_allocator)


#if defined(unix) || \
    defined(__unix__)

#include <unixstl.h>

#include <unixstl_performance_counter.h>

unixstl_ns_using(performance_counter)

#elif defined(WIN32) || \
    defined(_WIN32)

#include <winstl.h>

#include <winstl_performance_counter.h>

winstl_ns_using(performance_counter)

#else /* ? operating system */

# error Operating system not discriminated

#endif /* operating system */

#include <vector>

stlsoft_ns_using_std(vector)

/* /////////////////////////////////////////////////////////////////////////////
 * printf() format string
 */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(_INTEL_COMPILER) || \
    defined(_MSC_VER)
 #define _sint64_fmt    "%I64d"
 #define _uint64_fmt    "%I64u"
#elif defined(__DMC__) || \
      defined(__MWERKS__)
 #define _sint64_fmt    "%lld"
 #define _uint64_fmt    "%llu"
#else
 #error Further compiler discrimination is required
#endif /* compiler */

/* ////////////////////////////////////////////////////////////////////////// */

const size_t	INTERNAL_SIZE	=	96;

typedef int													element_t;
typedef	processheap_allocator<element_t>					int_ator_t;
typedef vector<element_t, int_ator_t>						int_vector_t;
typedef pod_vector<element_t, int_ator_t, INTERNAL_SIZE>	int_pvector_t;

/* ////////////////////////////////////////////////////////////////////////// */

template <class V>
unsigned long test_push_front(int iterations, int num_insertions)
{
	unsigned long	ret	=	0;

	for(int i = 0; i < iterations; ++i)
	{
//		int	volatile	one	=	1;
		V				v;
//		int	volatile	two	=	2;

		for(int j = 0; j < num_insertions; ++j)
		{
			v.insert(v.begin(), j);

#if 0
			if(1 != one)
			{
				printf("one!!\n");
			}
			if(2 != two)
			{
				printf("two!!\n");
			}
#endif /* 0 */

//			ret += v[0];
			ret += j;
		}
	}

	return ret;
}

template <class V>
unsigned long test_push_back(int iterations, int num_insertions)
{
	unsigned long	ret	=	0;

	for(int i = 0; i < iterations; ++i)
	{
		V	v;

		for(int j = 0; j < num_insertions; ++j)
		{
			v.insert(v.end(), j);

//			ret += v[0];
			ret += j;
		}
	}

	return ret;
}

template <class V>
unsigned long test_push_back_reserve(int iterations, int num_insertions)
{
	unsigned long	ret	=	0;

	for(int i = 0; i < iterations; ++i)
	{
		V	v;

		v.reserve(num_insertions);

		for(int j = 0; j < num_insertions; ++j)
		{
			v.insert(v.end(), j);

//			ret += v[0];
			ret += j;
		}
	}

	return ret;
}

template <class V>
unsigned long test_erase_front(int iterations, int num_erasures)
{
	unsigned long	ret	=	0;

	for(int i = 0; i < iterations; ++i)
	{
		V	v(num_erasures);

		for(int j = 0; j < num_erasures; ++j)
		{
			v.erase(v.begin());

//			ret += v[0];
			ret += j;
		}
	}

	return ret;
}

template <class V>
unsigned long test_erase_back(int iterations, int num_erasures)
{
	unsigned long	ret	=	0;

	for(int i = 0; i < iterations; ++i)
	{
		V	v(num_erasures);

		for(int j = 0; j < num_erasures; ++j)
		{
#if defined(__STLSOFT_COMPILER_IS_GCC) || \
    defined(__STLSOFT_COMPILER_IS_MWERKS)
			ss_typename_type_k V::iterator	it = v.begin();

			stlsoft_ns_qual_std(advance)(it, v.size() - 1);
#else /* ? 0 */
			ss_typename_type_k V::iterator	it = &v[v.size() - 1];
#endif /* 0 */

			v.erase(it);

//			ret += v[0];
			ret += j;
		}
	}

	return ret;
}

template <class V>
unsigned long test_insert_block_back(int iterations, int num_insertions)
{
	unsigned long	ret	=	0;

	V		ints_;
	V const	&ints	=	ints_;

	ints_.push_back(1);
	ints_.push_back(2);
	ints_.push_back(3);
	ints_.push_back(4);

	for(int i = 0; i < iterations; ++i)
	{
		V	v;

		for(int j = 0; j < num_insertions; ++j)
		{
			v.insert(v.end(), ints.begin(), ints.end());

//			ret += v[0];
			ret += j;
		}
	}

	return ret;
}



/* ////////////////////////////////////////////////////////////////////////// */

int main(int argc, char *argv[])
{
#if 0
	::Sleep(100000);
#endif /* 0 */

    int                 iRet			=   0;
    performance_counter counter;
    int                 i;
#ifdef _DEBUG
	int					ITERATIONS		=	10;
#else /* ? _DEBUG */
	int					ITERATIONS		=	50000;
#endif /* _DEBUG */
//	int					NUM_INSERTIONS	=	1000;

#ifndef SUCCINCT_OUTPUT_DDJ_0504
	puts("pod_vector_speed_test: " __STLSOFT_COMPILER_LABEL_STRING);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */

	for(i = 1; i < argc; ++i)
	{
		const char	*arg	=	argv[i];

		if(arg[0] == '-')
		{
			if(arg[1] == '-')
			{
				/* -- arguments */
			}
			else
			{
				/* - arguments */
			}
		}
		else
		{
			/* other arguments */
		}
	}

#ifdef WIN32
	::SetPriorityClass(::GetCurrentProcess(), HIGH_PRIORITY_CLASS);
	::SetThreadPriority(::GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
#endif // WIN32

	/* . */
#ifdef _DEBUG
	test_push_front<int_pvector_t>(1, 100);
	test_push_back<int_pvector_t>(1, 100);
	test_push_back_reserve<int_pvector_t>(1, 100);
	test_erase_front<int_pvector_t>(1, 100);
	test_erase_back<int_pvector_t>(1, 100);
	test_insert_block_back<int_pvector_t>(1, 100);
#endif /* _DEBUG */

	int	NUM_INC	=	50;
	int	NUM_MIN	=	50;
	int	NUM_MAX	=	150;

	performance_counter::interval_type	totalVector		=	0;
	performance_counter::interval_type	totalPODVector	=	0;

#if 0
	try
	{
#endif /* 0 */
		for(int WARMUPS = 2; WARMUPS > 0; --WARMUPS)
		{
#if 1
			for(int num_push_front = NUM_MIN; num_push_front < NUM_MAX; num_push_front += NUM_INC)
			{
				counter.start();
				test_push_front<int_vector_t>(ITERATIONS, num_push_front);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("std::vector, push-front (%d) \t%lu\n", num_push_front, totalVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
		
				counter.start();
				test_push_front<int_pvector_t>(ITERATIONS, num_push_front);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalPODVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("stlsoft::pod_vector, push-front (%d) \t%lu\n", num_push_front, totalPODVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}

				if(1 == WARMUPS)
				{
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("\n");
					printf("\n");
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
			}
#endif /* 0 */
		
			for(int num_push_back = NUM_MIN; num_push_back < NUM_MAX; num_push_back += NUM_INC)
			{
				counter.start();
				test_push_back<int_vector_t>(ITERATIONS, num_push_back);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("std::vector, push-back (%d) \t%lu\n", num_push_back, totalVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
		
				counter.start();
				test_push_back<int_pvector_t>(ITERATIONS, num_push_back);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalPODVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("stlsoft::pod_vector, push-back (%d) \t%lu\n", num_push_back, totalPODVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}

				if(1 == WARMUPS)
				{
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("\n");
					printf("\n");
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
			}
		
			for(int num_push_back_reserve = NUM_MIN; num_push_back_reserve < NUM_MAX; num_push_back_reserve += NUM_INC)
			{
				counter.start();
				test_push_back_reserve<int_vector_t>(ITERATIONS, num_push_back_reserve);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("std::vector, push-back-reserve (%d) \t%lu\n", num_push_back_reserve, totalVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
		
				counter.start();
				test_push_back_reserve<int_pvector_t>(ITERATIONS, num_push_back_reserve);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalPODVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("stlsoft::pod_vector, push-back-reserve (%d) \t%lu\n", num_push_back_reserve, totalPODVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}

				if(1 == WARMUPS)
				{
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("\n");
					printf("\n");
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
			}
		
			for(int num_erasures_front = NUM_MIN; num_erasures_front < NUM_MAX; num_erasures_front += NUM_INC)
			{
				counter.start();
				test_erase_front<int_vector_t>(ITERATIONS, num_erasures_front);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("std::vector, erase-front (%d) \t%lu\n", num_erasures_front, totalVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
		
				counter.start();
				test_erase_front<int_pvector_t>(ITERATIONS, num_erasures_front);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalPODVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("stlsoft::pod_vector, erase-front (%d) \t%lu\n", num_erasures_front, totalPODVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}

				if(1 == WARMUPS)
				{
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("\n");
					printf("\n");
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
			}
		
			for(int num_erasures_back = NUM_MIN; num_erasures_back < NUM_MAX; num_erasures_back += NUM_INC)
			{
				counter.start();
				test_erase_back<int_vector_t>(ITERATIONS, num_erasures_back);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("std::vector, erase-back (%d) \t%lu\n", num_erasures_back, totalVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
		
				counter.start();
				test_erase_back<int_pvector_t>(ITERATIONS, num_erasures_back);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalPODVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("stlsoft::pod_vector, erase-back (%d) \t%lu\n", num_erasures_back, totalPODVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}

				if(1 == WARMUPS)
				{
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("\n");
					printf("\n");
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
			}
		
			for(int num_block_insertions_back = NUM_MIN; num_block_insertions_back < NUM_MAX; num_block_insertions_back += NUM_INC)
			{
				counter.start();
				test_insert_block_back<int_vector_t>(ITERATIONS, num_block_insertions_back);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("std::vector, insert-block-back (%d) \t%lu\n", num_block_insertions_back, totalVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
		
				counter.start();
				test_insert_block_back<int_pvector_t>(ITERATIONS, num_block_insertions_back);
				counter.stop();
				if(1 == WARMUPS)
				{
					totalPODVector += counter.get_milliseconds();
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("stlsoft::pod_vector, insert-block-back (%d) \t%lu\n", num_block_insertions_back, totalPODVector);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}

				if(1 == WARMUPS)
				{
#ifndef SUCCINCT_OUTPUT_DDJ_0504
					printf("\n");
					printf("\n");
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */
				}
			}
		}
#if 0
	}
	catch(stlsoft_ns_qual_std(bad_alloc) &x)
	{
		puts(x.what());
	}
#endif /* 0 */

#ifndef SUCCINCT_OUTPUT_DDJ_0504
	printf("std::vector total\t%lu\n", totalVector);
	printf("stlsoft::pod_vector total\t%lu\n", totalPODVector);
#else /* ? SUCCINCT_OUTPUT_DDJ_0504 */
	fprintf(stdout,
			"pod_vector (%s):\t" _sint64_fmt " in %d repeats\n",
			__STLSOFT_COMPILER_LABEL_STRING,
			totalPODVector,
			ITERATIONS);
#endif /* !SUCCINCT_OUTPUT_DDJ_0504 */


	STLSOFT_SUPPRESS_UNUSED(NUM_INC);
	STLSOFT_SUPPRESS_UNUSED(NUM_MIN);
	STLSOFT_SUPPRESS_UNUSED(NUM_MAX);

    return iRet;
}

/* ////////////////////////////////////////////////////////////////////////// */
