/* /////////////////////////////////////////////////////////////////////////////
 * File:		stringtok.cpp
 *
 * Purpose:		Test program for string tokenisation
 *
 * Created		16th June 2003
 * Updated:		21st February 2004
 *
 * Copyright:	Synesis Software Pty Ltd, (c) 2003. All rights reserved.
 *
 * Home:		www.synesis.com.au/software
 *
 * ////////////////////////////////////////////////////////////////////////// */


#include <stdio.h>

#include <stlsoft.h>

#if defined(STRINGTOK_BOOST)
# include <boost/tokenizer.hpp>
#elif defined(STRINGTOK_STLSOFT)
# include <stlsoft_string_tokeniser.h>
#else
# error Must specify one of STRINGTOK_BOOST or STRINGTOK_STLSOFT
#endif /* 0 */

#include <winstl_processheap_allocator.h>

#include <stlsoft_simple_string.h>

typedef stlsoft_ns_qual(basic_simple_string)<	char
											,	stlsoft_ns_qual(char_traits)<char>
											,	winstl_ns_qual(processheap_allocator)<char>
											>			string_t;

#if defined(WIN32) || \
    defined(_WIN32)
# include <winstl_performance_counter.h>

 typedef winstl_ns_qual(performance_counter)			performance_counter_t;
#else
# include <unixstl_performance_counter.h>

 typedef unixstl_ns_qual(performance_counter)			performance_counter_t;
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * printf() format string
 */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(_INTEL_COMPILER) || \
    defined(_MSC_VER)
# define _sint64_fmt    "%I64d"
# define _uint64_fmt    "%I64u"
#elif defined(__DMC__) || \
      defined(__MWERKS__)
# define _sint64_fmt    "%lld"
# define _uint64_fmt    "%llu"
#else
# error Further compiler discrimination is required
#endif /* compiler */

/* ////////////////////////////////////////////////////////////////////////// */

const int	WARMUPS			=	2;
const int	DEF_NUM_STRINGS	=	100;
const int	DEF_ITERATIONS	=	100;

/* ////////////////////////////////////////////////////////////////////////// */

/* Avoid the painful mess that is compiling with max & Windows headers */
inline int int_max(int i1, int i2)
{
	return i1 < i2 ? i2 : i1;
}

/* ////////////////////////////////////////////////////////////////////////// */

int main(int argc, char **argv)
{
	performance_counter_t	counter;
	int						NUM_STRINGS	=	(argc > 1) ? atoi(argv[1]) : 0;
	int						ITERATIONS	=	(argc > 2) ? atoi(argv[2]) : 0;
	int						w;
	int						i			=	0;
	int						len			=	0;
	string_t				src			=	";";

	NUM_STRINGS	=	int_max(NUM_STRINGS, DEF_NUM_STRINGS);
	ITERATIONS	=	int_max(ITERATIONS, DEF_ITERATIONS);

	for(i = 0; i < NUM_STRINGS; ++i)
	{
		char	temp[201] = " ";

//printf("%d: %*lu;\n", i, (i & 0x1f), ::GetVersion() & 0x7fff);
		sprintf(temp, "%*lu;", (i & 0x1f), ::GetVersion() & 0x7fff);
		src += temp;
//printf("%d: %s\n", i, src.c_str());
	}

//printf("Exiting\n");
//return 0;

#ifdef WIN32
	::SetPriorityClass(::GetCurrentProcess(), HIGH_PRIORITY_CLASS);
	::SetThreadPriority(::GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
#endif // WIN32

	for(w = 0; w < WARMUPS; ++w)
	{
		counter.start();

		for(i = 0, len = 0; i < ITERATIONS; ++i)
		{
#if defined(STRINGTOK_BOOST)
			typedef boost::tokenizer<boost::char_separator<char>, string_t::const_iterator, string_t>	tokeniser_t;

			boost::char_separator<char>		sep(";");
			tokeniser_t						tokens(src, sep);
#elif defined(STRINGTOK_STLSOFT)
			typedef stlsoft::string_tokeniser<string_t, char>	tokeniser_t;

			tokeniser_t						tokens(src, ';');
#endif /* 0 */

			for (tokeniser_t::const_iterator tok_iter = tokens.begin(); tok_iter != tokens.end(); ++tok_iter)
			{
				len += (*tok_iter).size();
			}
		}

		counter.stop();
	}

	fprintf(stdout,
#if defined(STRINGTOK_BOOST)
			"stringtok-Boost (%s):\t" _sint64_fmt " in %d repeats, %d strings\n",
#elif defined(STRINGTOK_STLSOFT)
			"stringtok-STLSoft (%s):\t" _sint64_fmt " in %d repeats, %d strings\n",
#endif /* 0 */
			__STLSOFT_COMPILER_LABEL_STRING,
			counter.get_milliseconds(),
			ITERATIONS,
			NUM_STRINGS);

	return len;
}

/* ////////////////////////////////////////////////////////////////////////// */
